import Std
import Aesop

namespace WithAesop

@[simp]
def Eqlist (xs ys : List α) : Prop :=
  ∀ x, x ∈ xs ↔ x ∈ ys

def Eqlist.refl : Eqlist xs xs := by
  aesop

def Eqlist.trans : Eqlist xs ys → Eqlist ys zs → Eqlist xs zs := by
  aesop

def Eqlist.symm : Eqlist xs ys → Eqlist ys xs := by
  aesop

def Eqlist.equivalence : Equivalence (@Eqlist α) where
  refl _ := Eqlist.refl
  symm := Eqlist.symm
  trans := Eqlist.trans

instance eqlist : Setoid (List α) where
  r := Eqlist
  iseqv := Eqlist.equivalence

@[aesop unsafe apply]
theorem Eqlist.r_t_l {xs ys : List α} : x ∈ xs → ys ≈ xs → x ∈ ys := by
  intro m h
  have := h x
  aesop

@[aesop unsafe apply]
theorem Eqlist.l_t_r {xs ys : List α} : x ∈ xs → xs ≈ ys → x ∈ ys := by
  intro m h
  have := h x
  aesop

def Set α := Quotient (@eqlist α)

@[aesop simp]
def Set.fromList (xs : List α) : Set α := Quotient.mk _ xs

macro "{" e:term,* "}" : term => ``(Set.fromList [ $e,* ])

def someEvens : Set Nat := {2, 4, 6}

attribute [aesop simp] HasEquiv.Equiv
attribute [aesop simp] Setoid.r

example : someEvens = {4, 2, 6, 2} := by
  apply Quotient.sound
  aesop


def oneTwoThree : Set Nat := .fromList [1, 2, 3]

def Set.Mem (x : α) (xs : Set α) : Prop :=
  xs.lift (x ∈ ·) resp
where
  resp xs ys : xs ≈ ys → (x ∈ xs) = (x ∈ ys) := by
    aesop

instance : Membership α (Set α) where
  mem := Set.Mem

theorem falseEq (h1 h2 : ¬p) : isFalse h1 = isFalse h2 := by
  aesop

attribute [aesop unsafe forward] Decidable.isTrue
attribute [aesop unsafe forward] Decidable.isFalse

def Set.Mem.dec [DecidableEq α] (x : α) (xs : Set α) : Decidable (x ∈ xs) :=
  xs.hrecOn doIt <| by
    intros
    apply Subsingleton.helim
    congr 2
    apply Quotient.sound
    assumption
where
  doIt (xs : List α) : Decidable (Set.Mem x (Quotient.mk eqlist xs)) :=
    let h : Decidable (x ∈ xs) := inferInstance
    by aesop

instance [DecidableEq α] {xs : Set α} : Decidable (x ∈ xs) := Set.Mem.dec _ _

#eval if 3 ∈ someEvens then "yay" else "nope"



def Set.union (xs ys : Set α) : Set α :=
  xs.liftOn₂ ys (Set.fromList <| · ++ ·) <| by
    intros
    apply Quotient.sound
    aesop
