import Lean
import Std

inductive T where
  | bool
  | arr (dom ran : T)

infixr:30 " ⇒ " => T.arr

inductive Context where
  | nil
  | extend (Γ : Context) (x : String) (t : T)

notation "⬝" => Context.nil
notation:max Γ ";" x:1 " : " t:100 => Context.extend Γ x t


inductive InContext : Context → String → T → Type where
  | last : {Γ : Context} → InContext (Γ; x : t) x t
  | step : InContext Γ x t → InContext (Γ; y : t') x t

notation:20 Γ "(" x ")" " = " t:40 => InContext Γ x t

inductive Expr : Context → T → Type where
  | var (x : String) (ok : Γ(x) = t) : Expr Γ t
  | lam (x : String) (body : Expr (Γ; x : t) t') : Expr Γ (t ⇒ t')
  | app (e1 : Expr Γ (t ⇒ t')) (e2 : Expr Γ t) : Expr Γ t'
  | cstB (b : Bool) : Expr Γ .bool
  | «if» (c : Expr Γ .bool) (y n : Expr Γ t) : Expr Γ t

abbrev T.interp : T → Type
  | .bool => Bool
  | a ⇒ b => a.interp → b.interp

notation "⟦" t "⟧" => T.interp t

inductive Env : Context → Type where
  | nil : Env .nil
  | cons : ⟦t⟧ → Env Γ → Env (Γ; x : t)

notation:1023 ρ "[" y "↦" v "]" => Env.cons (x := y) v ρ

def Env.lookup (ρ : Env Γ) (var : InContext Γ x t) : ⟦t⟧ :=
  match ρ, var with
  | .cons v _, .last => v
  | .cons _ vs, .step var' => vs.lookup var'

notation ρ "(" x ")" => Env.lookup ρ x

def Expr.eval (ρ : Env Γ) : Expr Γ t → ⟦t⟧
  | .var _ ok => ρ(ok)
  | .lam x body => fun v => body.eval ρ[x ↦ v]
  | .app e1 e2 => e1.eval ρ (e2.eval ρ)
  | .cstB b => b
  | .«if» c y n =>
    if c.eval ρ then y.eval ρ else n.eval ρ

declare_syntax_cat lam
syntax ident : lam
syntax "(" lam  ")" : lam
syntax "fun " "(" ident ":" term ")" " => " lam : lam
syntax:48 lam:50 lam:48 : lam
syntax &"true" : lam
syntax &"false" : lam
syntax "if " lam " then " lam " else " lam : lam

syntax "{{" lam "}}" : term

open Lean

macro_rules
  | `({{ ($e) }}) => pure e
  | `({{ true }}) => `(Expr.cstB true)
  | `({{ false }}) => `(Expr.cstB false)
  | `({{ $x:ident }}) => `(Expr.var $(quote x.getId.toString) (by repeat constructor))
  | `({{ fun ($x : $t) => $body }}) => `(Expr.lam (t := $t) $(quote x.getId.toString) {{ $body }})
  | `({{ $e1 $e2 }}) => `(Expr.app {{$e1}} {{$e2}})
  | `({{ if $e1 then $e2 else $e3}}) => `(Expr.«if» {{$e1}} {{$e2}} {{$e3}})



#eval Expr.eval (Γ := ⬝; "x" : .bool) .nil["x" ↦ true] {{ x }}
#eval Expr.eval .nil {{ fun (x : .bool) => x }} <| true
#eval Expr.eval .nil {{ fun (x : .bool ⇒ .bool) => x false }} <| not
#eval Expr.eval .nil {{ fun (x : .bool) => if x then false else true  }} <| true

  