import Std

def Eqlist (xs ys : List α) : Prop :=
  ∀ x, x ∈ xs ↔ x ∈ ys

def Eqlist.refl : Eqlist xs xs := by
  intro
  rfl

def Eqlist.trans : Eqlist xs ys → Eqlist ys zs → Eqlist xs zs := by
  intro h1 h2
  intro x
  constructor <;> intro hMem
  . apply (h2 x).mp
    apply (h1 x).mp
    assumption
  . apply (h1 x).mpr
    apply (h2 x).mpr
    assumption

def Eqlist.symm : Eqlist xs ys → Eqlist ys xs := by
  intro h
  unfold Eqlist
  unfold Eqlist at h
  simp [*]

def Eqlist.equivalence : Equivalence (@Eqlist α) where
  refl _ := Eqlist.refl
  symm := Eqlist.symm
  trans := Eqlist.trans

instance eqlist : Setoid (List α) where
  r := Eqlist
  iseqv := Eqlist.equivalence

theorem Eqlist.r_t_l {xs ys : List α} : x ∈ xs → ys ≈ xs → x ∈ ys := by
  intro mem h
  have iff := h x
  apply iff.mpr; assumption

theorem Eqlist.l_t_r {xs ys : List α} : x ∈ xs → xs ≈ ys → x ∈ ys := by
  intro mem h
  have := h x
  apply this.mp; assumption


example : [2, 4, 6] ≈ [6, 2, 4, 2, 6] := by sorry


-------------------------------------

def Set α := Quotient (@eqlist α)

def Set.fromList (xs : List α) : Set α := Quotient.mk _ xs

macro "{" e:term,* "}" : term => ``(Set.fromList [ $e,* ])

def someEvens : Set Nat := {2, 4, 6}

example : someEvens = {4, 2, 6, 2} := by sorry

def Set.Mem (x : α) (xs : Set α) : Prop :=
  xs.lift (x ∈ ·) resp
where
  resp xs ys : xs ≈ ys → (x ∈ xs) = (x ∈ ys) := by sorry

instance : Membership α (Set α) where
  mem := Set.Mem

theorem falseEq (h1 h2 : ¬p) : isFalse h1 = isFalse h2 := by
  congr

def Set.Mem.dec [DecidableEq α] (x : α) (xs : Set α) : Decidable (x ∈ xs) :=
  xs.hrecOn doIt fun (xs ys : List α) (h : xs ≈ ys) => by sorry
where
  doIt (xs : List α) : Decidable (Set.Mem x (Quotient.mk eqlist xs)) :=
    if h : x ∈ xs then
      .isTrue h
    else
      .isFalse fun hMem => h hMem

instance [DecidableEq α] {xs : Set α} : Decidable (x ∈ xs) := Set.Mem.dec _ _

#eval if 3 ∈ someEvens then "yay" else "nope"


def Set.union (xs ys : Set α) : Set α :=
  xs.liftOn₂ ys (Set.fromList <| · ++ ·) <| by sorry
