
-- Comments start with two dashes, and go to the end
-- of the line.

/- Or they start with slash-dash,
and end with dash-slash -/


-- Instead of a REPL, Lean provides commands to check types and evaluate
-- expressions within the editor:
#eval 2 + 2

#check "hello!"

-- As in Haskell and ML, function calls are just spaces - no parentheses or commas:
#eval String.replace "2 < 5" "<" "&lt;"

#eval List.reverse [1, 2, 3, 4, 5]

-- Conditionals are expressions
#eval if 2 < 5 then "smaller" else "larger"

-- There is pervasive overloading (more on that later)
#check 2
#check -2
#check (2 : Int)
#check 2.0
#check (2 : Float)
#eval 2 == 2.0
#eval 2 + 1.5

-- Definitions are created with def
def five := 5

#check five
#eval five

def hello := "Hello, "

def world := "world!"

#eval hello ++ world

def fivePointOh : Float := 5

#eval fivePointOh

def MyString := String

def greeting : MyString := hello ++ world

-- Function arguments go before the :=

def add1 x := x + 1
#check add1
#eval add1 88

--def add x y := x + y

def add (x : Int) (y : Int) : Int := x + y

namespace MyNat

inductive Natural where
  | zero : Natural
  | succ : Natural → Natural
deriving Repr

--#check zero
#check Natural.zero
#eval Natural.zero

open Natural in
#check zero

def sub1 (n : Natural) : Natural :=
  match n with
  | .zero => .zero
  | .succ k => k

def add (n : Natural) (k : Natural) : Natural :=
  match k with
  | .zero => n
  | .succ k' => .succ (add n k')

def add' (n : Natural) : Natural → Natural
  | .zero => n
  | .succ k => .succ (add' n k)

end MyNat

namespace MyList

inductive List (α : Type) : Type where
  | nil : List α
  | cons : α → List α → List α



end MyList

structure Recipe where
  ingredients : List String
  instructions : List String
  /-- Preparation time in minutes -/
  prepTime : Nat
deriving Repr

def tea : Recipe := {
  ingredients := ["water", "tea leaves"],
  instructions := [
    "boil water",
    "add leaves",
    "wait"
  ],
  prepTime := 4
}

def sandwich : Recipe := ⟨["bread", "peanut butter", "jam"], ["put filling on bread", "close"], 2⟩

def appleSlices : Recipe where
  ingredients := ["apple"]
  instructions := ["cut apple"]
  prepTime := 1

def fastest (r1 r2 : Recipe) : Recipe :=
  if r1.prepTime < r2.prepTime then r1 else r2

def easiest (r1 r2 : Recipe) : Recipe :=
  match r1, r2 with
  | ⟨_, steps1, _⟩, ⟨_, steps2, _⟩ =>
    if List.length steps1 < List.length steps2 then r1 else r2

-- Property notation

#eval List.length [1, 2, 3, 4]
#eval [1, 2, 3, 4].length

#check List.length
#check Recipe.ingredients
