import Std

inductive Vec (α : Type u) : Nat → Type u where
  | nil : Vec α 0
  | cons : α → Vec α n → Vec α (n + 1)

-- When pattern matching with dependent types, one pattern affects the others

def Vec.map (f : α → β) (xs : Vec α n) : Vec β n :=
  match n, xs with
  | 0, .nil => .nil
  | n' + 1, .cons y ys => .cons (f y) (ys.map f)

def Vec.zip : Vec α n → Vec β n → Vec (α × β) n
  | nil, nil => nil
  | cons x xs, cons y ys => cons (x, y) (zip xs ys)


def Vec.zipWith (f : α → β → γ) : Vec α n → Vec β n → Vec γ n
  | nil, nil => nil
  | cons x xs, cons y ys => cons (f x y) (zipWith f xs ys)

-- Sometimes, we're not so lucky

-- def Nat.add : Nat → Nat → Nat
--   | a, Nat.zero   => a
--   | a, Nat.succ b => Nat.succ (Nat.add a b)



def Vec.append : Vec α n → Vec α k → Vec α (n + k)
  | nil, ys =>
    Eq.symm (Nat.zero_add k) ▸ ys
  | cons x xs, ys =>
    have : ∀ n, n + 1 + k = n + k + 1 := by
      intros; simp_arith
    this _ ▸ cons x (append xs ys)


def Vec.append' : Vec α n → Vec α k → Vec α (k + n)
  | nil, ys => ys
  | cons x xs, ys => cons x (append' xs ys)

structure Vec' (α : Type u) (n : Nat) where
  contents : List α
  prf : contents.length = n

def Vec'.append : Vec' α n → Vec' α k → Vec' α (k + n)
  | ⟨xs, prf1⟩, ⟨ys, prf2⟩ => ⟨xs ++ ys, by simp_arith [*]⟩


def Vec.lookup : Vec α n → Fin n → α
  | .nil, i => i.elim0
  | .cons x xs, ⟨0, _⟩ => x
  | .cons x xs, ⟨i+1, prf⟩ =>
    xs.lookup ⟨i, Nat.lt_of_succ_lt_succ prf⟩
